import configparser
import logging
import shutil
from copy import deepcopy
from pathlib import Path
from typing import Dict, Any, Union

import toml

from transfer_nlp.plugins.config import ExperimentConfig
from transfer_nlp.plugins.reporters import ReporterABC
from transfer_nlp.plugins.trainer_abc import TrainerABC

ConfigEnv = Dict[str, Any]


def load_config(p: Path) -> Dict[str, ConfigEnv]:
    p = Path(str(p)).expanduser()
    if p.suffix == '.toml':
        rv = toml.load(p)
        return rv

    if p.suffix != '.cfg':
        raise ValueError("Config files should be either .cfg or .toml files")

    def get_val(cfg: configparser.ConfigParser, section: str, key):
        try:
            return cfg.getint(section, key)
        except ValueError:
            pass
        try:
            return cfg.getfloat(section, key)
        except ValueError:
            pass
        try:
            return cfg.getboolean(section, key)
        except ValueError:
            pass

        return cfg[section][key]

    cfg = configparser.ConfigParser()
    cfg.optionxform = str
    cfg.read(p)

    rv = {}

    for exp_name in cfg.sections():
        exp = {}
        for key in cfg[exp_name].keys():
            exp[key] = get_val(cfg, exp_name, key)
        rv[exp_name] = exp

    return rv


class ExperimentRunner:
    """
    Run an experiment several times with varying configurations.

    This class facilitates reusing a single json experiment file across several different configuations.
    """

    @staticmethod
    def _capture_logs(report_path: Path):
        logger = logging.getLogger('')
        handler = logging.FileHandler(str(report_path / 'runner.log'))
        fmt = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')  # TODO configurable?
        handler.setFormatter(fmt)
        logger.addHandler(handler)
        return handler

    @staticmethod
    def _stop_log_capture(handler):
        logger = logging.getLogger('')
        logger.removeHandler(handler)

    @staticmethod
    def run_all(experiment: Union[str, Path],
                experiment_config: Union[str, Path],
                report_dir: Union[str, Path],
                trainer_config_name: str = 'trainer',
                reporter_config_name: str = 'reporter',
                experiment_cache: Union[str, Path, Dict] = None,
                **env_vars) -> ExperimentConfig:
        """
        :param experiment: the experiment config
        :param experiment_config: the experiment config file. The cfg file should be defined in `ConfigParser
               <https://docs.python.org/3/library/configparser.html#module-configparser>`_ format such that
               each section is an experiment configuration.
        :param report_dir: the directory in which to produce the reports. It's recommended to include a timestamp your report directory so you
               can preserve previous reports across code changes. E.g. $HOME/reports/run_2019_02_22.
        :param trainer_config_name: the name of the trainer configuration object. The referenced object should implement `TrainerABC`.
        :param reporter_config_name: the name of the reporter configuration object. The referenced object should implement `ReporterABC`.
        :param experiment_cache: the experiment config with cached objects
        :param env_vars: any additional environment variables, like file system paths
        :return: the experiment cache
        """

        envs: Dict[str, ConfigEnv] = load_config(Path(experiment_config))

        report_path = Path(report_dir)
        report_path.mkdir(parents=True)

        # Before starting, save the 3 global files: experiment, configs and cache
        global_report_dir = report_path / 'global-reporting'
        global_report_dir.mkdir(parents=True)
        shutil.copy(src=str(experiment), dst=str(global_report_dir / str(Path(experiment).name)))
        shutil.copy(src=str(experiment), dst=str(global_report_dir / str(Path(experiment_cache).name)))
        shutil.copy(src=str(experiment_config), dst=str(global_report_dir / str(Path(experiment_config).name)))

        experiment_config_cache = {}
        if experiment_cache:
            logging.info("#" * 5 + f"Building a set of read-only objects and cache them for use in different experiment settings" + "#" * 5)
            experiment_config_cache = ExperimentConfig(experiment_cache, **env_vars)
            logging.info("#" * 5 + f"Read-only objects are built and cached for use in different experiment settings" + "#" * 5)

        aggregate_reports = {}
        for exp_name, env in envs.items():
            exp_report_path = report_path / exp_name
            exp_report_path.mkdir()
            log_handler = ExperimentRunner._capture_logs(exp_report_path)
            try:
                logging.info('running %s', exp_name)
                all_vars = dict(env_vars)
                all_vars.update(env)

                exp = deepcopy(experiment)
                if experiment_cache:
                    exp = ExperimentConfig.load_experiment_config(exp)
                    exp.update(experiment_config_cache)

                experiment_config = ExperimentConfig(exp, **all_vars)
                trainer: TrainerABC = experiment_config[trainer_config_name]
                reporter: ReporterABC = experiment_config[reporter_config_name]
                trainer.train()

                # Save the config for this particular experiment
                exp_config = {
                    exp_name: all_vars}
                with (exp_report_path / 'experiment_config.toml').open('w') as expfile:
                    toml.dump(exp_config, expfile)

                # Get this particular config reporting and store it in the
                # aggregated reportings
                report = reporter.report(exp_name, experiment_config, exp_report_path)
                aggregate_reports[exp_name] = report
            finally:
                ExperimentRunner._stop_log_capture(log_handler)

        reporter_class = experiment_config[reporter_config_name].__class__
        if issubclass(reporter_class, ReporterABC):
            reporter_class.report_globally(aggregate_reports=aggregate_reports, report_dir=global_report_dir)

        return experiment_config_cache



"""
This class contains the abstraction interface to customize runners.
For the training loop, we use the engine logic from pytorch-ignite

Check experiments for examples of experiment json files

Examples using gradual unfreezing and adaptation methods in general are adapted from
the NAACL 2019 tutorial on TRansfer Learning for NLP https://colab.research.google.com/drive/1iDHCYIrWswIKp-n-pOg69xLoZO09MEgf#scrollTo=GObQkkttljWv&forceEdit=true&offline=true&sandboxMode=true
"""

import inspect
import logging
import re
from abc import abstractmethod
from collections import defaultdict
from typing import Dict, List, Any, Union, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from ignite.contrib.handlers.tensorboard_logger import TensorboardLogger, OutputHandler, OptimizerParamsHandler, WeightsScalarHandler, WeightsHistHandler, \
    GradsScalarHandler
from ignite.contrib.handlers.tqdm_logger import ProgressBar
from ignite.engine import Events
from ignite.engine.engine import Engine
from ignite.metrics import Loss, Metric, RunningAverage, MetricsLambda, Accuracy
from ignite.utils import convert_tensor

from transfer_nlp.loaders.loaders import DatasetSplits
from transfer_nlp.plugins.config import register_plugin
from transfer_nlp.plugins.regularizers import RegularizerABC
from transfer_nlp.plugins.trainer_abc import TrainerABC

logger = logging.getLogger(__name__)

# Tensorboard is used within PyTorch but is not a dependency, so it should be installed manually by users
TENSORBOARD = True
try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    logger.debug("To use torch.utils.tensorboard, please install tensorboard>=1.14, and future")
    TENSORBOARD = False


def set_seed_everywhere(seed: int, cuda: bool):
    np.random.seed(seed)
    torch.manual_seed(seed)
    if cuda:
        torch.cuda.manual_seed_all(seed)


def _prepare_batch(batch: Union[Dict, List, Tuple], device=None, non_blocking: bool = False):
    """Prepare batch for training: pass to a device with options.

    """
    if isinstance(batch, dict):
        result = {key: convert_tensor(value, device=device, non_blocking=non_blocking) for key, value in batch.items()}
        return result
    elif isinstance(batch, tuple):
        result = (convert_tensor(value, device=device, non_blocking=non_blocking) for value in batch)
        return result
    elif isinstance(batch, list):
        result = [convert_tensor(value, device=device, non_blocking=non_blocking) for value in batch]
        return result
    else:
        raise ValueError("Only dict, tuples and lists are valid for batch")


class TrainingMetric(Metric):

    def __init__(self, metric: Metric):
        self.source_metric = metric
        self.reset()

        super().__init__(lambda x: x[:-1])

    def reset(self):
        self.source_metric.reset()

    def update(self, output):

        if not isinstance(self.source_metric, MetricsLambda):
            self.source_metric.update(output)
            return

        # If a source metric is made of several metrics, e.g. MetricsLambda
        # metrics, we need to update each sub-metrics separately
        for source in self.source_metric.args:
            if isinstance(source, Metric):
                source.update(output)
        return

    def compute(self):
        return self.source_metric.compute()


@register_plugin
class BaseIgniteTrainer(TrainerABC):

    def __init__(self,
                 model: nn.Module,
                 dataset_splits: DatasetSplits,
                 loss: nn.Module,
                 optimizer: optim.Optimizer,
                 metrics: Dict[str, Metric],
                 device: str = None,
                 num_epochs: int = 1,
                 seed: int = None,
                 cuda: bool = None,
                 loss_accumulation_steps: int = 4,
                 scheduler: Any = None,
                 regularizer: RegularizerABC = None,
                 gradient_clipping: float = 1.0,
                 output_transform=None,
                 tensorboard_logs: str = None):

        self.model: nn.Module = model

        self.forward_param_defaults = {}

        model_spec = inspect.getfullargspec(model.forward)
        self.forward_params: List[str] = model_spec.args[1:]
        for fparam, pdefault in zip(reversed(model_spec.args[1:]), reversed(model_spec.defaults if model_spec.defaults else [])):
            self.forward_param_defaults[fparam] = pdefault

        self.dataset_splits: DatasetSplits = dataset_splits
        self.loss: nn.Module = loss
        self.optimizer: optim.Optimizer = optimizer
        self.metrics: Dict[str, Metric] = metrics
        self.metrics: List[Metric] = [metric for _, metric in self.metrics.items()]
        self.device: str = device
        self.num_epochs: int = num_epochs
        self.scheduler: Any = scheduler
        self.seed: int = seed
        self.cuda: bool = cuda
        if self.cuda is None:  # If cuda not specified, just check if the cuda is available and use accordingly
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.loss_accumulation_steps: int = loss_accumulation_steps
        self.regularizer: RegularizerABC = regularizer
        self.gradient_clipping: float = gradient_clipping
        self.output_transform = output_transform
        self.tensorboard_logs: str = tensorboard_logs
        if self.tensorboard_logs and TENSORBOARD:
            self.writer = SummaryWriter(log_dir=self.tensorboard_logs)

        if not self.output_transform:
            self.output_transform = lambda y_pred, y_target, loss: (y_pred, y_target, loss)

        self.eval_output_transform = lambda y, y_pred: (y, y_pred)

        self.trainer, self.training_metrics = self.create_supervised_trainer()
        self.evaluator = self.create_supervised_evaluator()

        loss_metrics = [m for m in self.metrics if isinstance(m, Loss)]

        if self.scheduler:
            if not loss_metrics:
                raise ValueError('A loss metric must be configured')
            elif len(loss_metrics) > 1:
                logging.warning('multiple loss metrics detected, using %s for LR scheduling', loss_metrics[0])
            self.loss_metric = loss_metrics[0]

        self.metrics_history = {
            "training": defaultdict(list),
            "validation": defaultdict(list),
            "test": defaultdict(list)}

        self.setup(self.training_metrics)

    def setup(self, training_metrics: Dict):
        def metric_name(n) -> str:
            if n.endswith('Accuracy'):
                n = 'acc'
            else:
                n = n[:-6] if n.endswith('Metric') else n
            return n

        def print_metrics(metrics) -> str:
            rv = ''
            metric_keys = sorted(k for k in metrics)
            for k in metric_keys:
                if k == 'Accuracy':
                    rv += f'{metric_name(k)}: {metrics[k]:.3} | '
                else:
                    rv += f'{metric_name(k)}: {metrics[k]} | '
            return rv

        def store_metrics(metrics: Dict, mode: str):
            metric_keys = sorted(k for k in metrics)
            for k in metric_keys:
                self.metrics_history[mode][metric_name(k)].append(metrics[k])

        if self.seed:
            set_seed_everywhere(self.seed, self.cuda)

        pbar = ProgressBar(persist=True)

        names = []
        for k, v in training_metrics.items():
            name = f'r{k}'
            names.append(name)
            RunningAverage(v).attach(self.trainer, name)
        RunningAverage(None, output_transform=lambda x: x[-1]).attach(self.trainer, 'rloss')

        names.append('rloss')
        pbar.attach(self.trainer, names)

        ProgressBar(persist=True).attach(engine=self.evaluator, metric_names=names)

        # A few events handler. To add / modify the events handler, you need to extend the __init__ method of RunnerABC
        # Ignite provides the necessary abstractions and a furnished repository of useful tools

        @self.trainer.on(Events.EPOCH_COMPLETED)
        def log_training_validation_results(trainer):

            self.evaluator.run(self.dataset_splits.train_data_loader())
            metrics = self.evaluator.state.metrics
            store_metrics(metrics=metrics, mode="training")
            logger.info(f"Training Results - Epoch: {trainer.state.epoch} {print_metrics(metrics)}")

            self.evaluator.run(self.dataset_splits.val_data_loader())
            metrics = self.evaluator.state.metrics
            store_metrics(metrics=metrics, mode="validation")
            logger.info(f"Validation Results - Epoch: {trainer.state.epoch} {print_metrics(metrics)}")

            metrics = self.trainer.state.metrics
            if self.scheduler:
                self.scheduler.step(metrics["rloss"])
                # self.scheduler.step(metrics[self.loss_metric.__class__.__name__])

        @self.trainer.on(Events.COMPLETED)
        def log_test_results(trainer):
            if self.dataset_splits.test_set:
                self.evaluator.run(self.dataset_splits.test_data_loader())
                metrics = self.evaluator.state.metrics
                store_metrics(metrics=metrics, mode="test")
                logger.info(f"Test Results - Epoch: {trainer.state.epoch} {print_metrics(metrics)}")

    def _forward(self, batch):
        model_inputs = {}
        for p in self.forward_params:
            val = batch.get(p)
            if val is None:
                if p in self.forward_param_defaults:
                    val = self.forward_param_defaults[p]
                else:
                    raise ValueError(f'missing model parameter "{p}"')

            model_inputs[p] = val

        return self.model(**model_inputs)

    @abstractmethod
    def update_engine(self, engine, batch):
        raise NotImplementedError

    @abstractmethod
    def infer_engine(self, engine, batch):
        raise NotImplementedError

    def create_supervised_trainer(self):

        if self.device:
            self.model.to(self.device)

        engine = Engine(self.update_engine)
        metrics = {}
        for i, metric in enumerate(self.metrics):
            if not isinstance(metric, Loss):
                n = metric.__class__.__name__
                tm = TrainingMetric(metric)
                metrics[n] = tm
                tm.attach(engine, n)

        return engine, metrics

    def create_supervised_evaluator(self):

        if self.device:
            self.model.to(self.device)

        engine = Engine(self.infer_engine)

        for i, metric in enumerate(self.metrics):
            metric.attach(engine, f'{str(metric.__class__.__name__)}')

        return engine

    def train(self):
        raise NotImplementedError


@register_plugin
class SingleTaskTrainer(BaseIgniteTrainer):

    def __init__(self,
                 model: nn.Module,
                 dataset_splits: DatasetSplits,
                 loss: nn.Module,
                 optimizer: optim.Optimizer,
                 metrics: Dict[str, Metric],
                 device: str = None,
                 num_epochs: int = 1,
                 seed: int = None,
                 cuda: bool = None,
                 loss_accumulation_steps: int = 4,
                 scheduler: Any = None,
                 regularizer: RegularizerABC = None,
                 gradient_clipping: float = 1.0,
                 output_transform=None,
                 tensorboard_logs: str = None,
                 optional_tensorboard_features: bool = False,
                 embeddings_name: str = None):

        super().__init__(
            model=model,
            dataset_splits=dataset_splits,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            device=device,
            num_epochs=num_epochs,
            seed=seed,
            cuda=cuda,
            loss_accumulation_steps=loss_accumulation_steps,
            scheduler=scheduler,
            regularizer=regularizer,
            gradient_clipping=gradient_clipping,
            output_transform=output_transform,
            tensorboard_logs=tensorboard_logs)

        self.optional_tensorboard_features: bool = optional_tensorboard_features
        self.embeddings_name: str = embeddings_name

        self.custom_setup()

    def custom_setup(self):

        if self.tensorboard_logs:
            tb_logger = TensorboardLogger(log_dir=self.tensorboard_logs)
            tb_logger.attach(self.trainer,
                             log_handler=OutputHandler(tag="training", output_transform=lambda loss: {
                                 'loss': loss}),
                             event_name=Events.ITERATION_COMPLETED)
            tb_logger.attach(self.evaluator,
                             log_handler=OutputHandler(tag="validation",
                                                       metric_names=["LossMetric"],
                                                       another_engine=self.trainer),
                             event_name=Events.EPOCH_COMPLETED)

            if self.optional_tensorboard_features:
                tb_logger.attach(self.trainer,
                                 log_handler=OptimizerParamsHandler(self.optimizer),
                                 event_name=Events.ITERATION_STARTED)
                tb_logger.attach(self.trainer,
                                 log_handler=WeightsScalarHandler(self.model),
                                 event_name=Events.ITERATION_COMPLETED)
                tb_logger.attach(self.trainer,
                                 log_handler=WeightsHistHandler(self.model),
                                 event_name=Events.EPOCH_COMPLETED)
                tb_logger.attach(self.trainer,
                                 log_handler=GradsScalarHandler(self.model),
                                 event_name=Events.ITERATION_COMPLETED)

            # This is important to close the tensorboard file logger
            @self.trainer.on(Events.COMPLETED)
            def end_tensorboard(trainer):
                logger.info("Training completed")
                tb_logger.close()

        if self.embeddings_name:
            @self.trainer.on(Events.COMPLETED)
            def log_embeddings(trainer):
                if hasattr(self.model, self.embeddings_name) and hasattr(self.dataset_splits, "vectorizer") and TENSORBOARD:
                    logger.info(f"Logging embeddings ({self.embeddings_name}) to Tensorboard!")
                    embeddings = getattr(self.model, self.embeddings_name).weight.data
                    metadata = [str(self.dataset_splits.vectorizer.data_vocab._id2token[token_index]).encode('utf-8') for token_index in
                                range(embeddings.shape[0])]
                    self.writer.add_embedding(mat=embeddings, metadata=metadata, global_step=self.trainer.state.epoch)

    def update_engine(self, engine, batch):

        # Gradient accumulation trick adapted from :
        # https://medium.com/huggingface/training-larger-batches-practical-tips-on-1-gpu-multi-gpu-distributed-setups-ec88c3e51255

        self.model.train()
        batch = _prepare_batch(batch, device=self.device, non_blocking=False)
        if isinstance(batch, dict):
            y_pred = self._forward(batch)
            loss = self.loss(input=y_pred, target=batch['y_target'])
        elif isinstance(batch, tuple) or isinstance(batch, list):
            y_pred = self.model.forward(*batch[:-1])
            loss = self.loss(input=y_pred, target=batch[-1])
        else:
            raise ValueError("Only dict, tuples and lists are valid for batch")

        # Add a regularisation term at train time only
        if self.regularizer:
            loss += self.regularizer.compute_penalty(model=self.model)

        loss /= self.loss_accumulation_steps

        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.gradient_clipping)

        if engine.state.iteration % self.loss_accumulation_steps == 0:
            self.optimizer.step()
            self.optimizer.zero_grad()
        if isinstance(batch, dict):
            return self.output_transform(y_pred, batch['y_target'], loss.item())
        elif isinstance(batch, tuple) or isinstance(batch, list):
            return self.output_transform(y_pred, batch[-1], loss.item())
        else:
            raise ValueError("Only dict, tuples and lists are valid for batch")

    def infer_engine(self, engine, batch):

        self.model.eval()
        with torch.no_grad():
            batch = _prepare_batch(batch, device=self.device, non_blocking=False)
            if isinstance(batch, dict):
                y_pred = self._forward(batch)
                return self.eval_output_transform(y_pred, batch['y_target'])
            elif isinstance(batch, tuple) or isinstance(batch, list):
                y_pred = self.model.forward(*batch[:-1])
                return self.eval_output_transform(y_pred, batch[-1])
            else:
                raise ValueError("Only dict, tuples and lists are valid for batch")

    def train(self):
        """
        Launch the ignite training pipeline
        If fine-tuning mode is granted in the config file, freeze all layers, replace classification layer by a Linear layer
        and reset the optimizer
        :return:
        """

        self.trainer.run(self.dataset_splits.train_data_loader(), max_epochs=self.num_epochs)


@register_plugin
class SingleTaskFineTuner(SingleTaskTrainer):

    def __init__(self,
                 model: nn.Module,
                 dataset_splits: DatasetSplits,
                 loss: nn.Module,
                 optimizer: optim.Optimizer,
                 metrics: Dict[str, Metric],
                 device: str = None,
                 num_epochs: int = 1,
                 seed: int = None,
                 cuda: bool = None,
                 loss_accumulation_steps: int = 4,
                 scheduler: Any = None,
                 regularizer: RegularizerABC = None,
                 gradient_clipping: float = 1.0,
                 output_transform=None,
                 tensorboard_logs: str = None,
                 optional_tensorboard_features: bool = False,
                 embeddings_name: str = None,
                 adaptation: str = 'hard-freezing',
                 decreasing_factor: int = 2.6,
                 pretrained: bool = False):
        super().__init__(
            model=model,
            dataset_splits=dataset_splits,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            device=device,
            num_epochs=num_epochs,
            seed=seed,
            cuda=cuda,
            loss_accumulation_steps=loss_accumulation_steps,
            scheduler=scheduler,
            regularizer=regularizer,
            gradient_clipping=gradient_clipping,
            output_transform=output_transform,
            tensorboard_logs=tensorboard_logs,
            optional_tensorboard_features=optional_tensorboard_features,
            embeddings_name=embeddings_name
        )
        self.adaptation: str = adaptation
        self.decreasing_factor: int = decreasing_factor
        self.pretrained: bool = pretrained

    def load_pretrained_model(self):
        """
        This methid is not implemented so that pytorch_pretrained_bert is not a 
        required dependency. Use these lines to implement the method if using
        pytorch_pretrained_bert
        Returns:

        """
        # from pytorch_pretrained_bert import cached_path
        # logger.info("Loading pretrained model")
        # state_dict = torch.load(cached_path("https://s3.amazonaws.com/models.huggingface.co/"
        #                                     "naacl-2019-tutorial/model_checkpoint.pth"), map_location=self.device)
        # self.model.load_state_dict(state_dict, strict=False)
        # logger.info("Pretrained model loaded!")

        raise NotImplementedError

    def freeze_params(self):

        for name, param in self.model.named_parameters():
            if 'embeddings' not in name and 'classification' not in name and 'adapters_1' not in name and 'adapters_2' not in name:
                param.detach_()
                param.requires_grad = False

            else:
                param.requires_grad = True

        full_parameters = sum(p.numel() for p in self.model.parameters())
        trained_parameters = sum(p.numel() for p in self.model.parameters() if p.requires_grad)

        logger.info(f"We will train {trained_parameters:3e} parameters out of {full_parameters:3e},"
                    f" i.e. {100 * trained_parameters / full_parameters:.2f}%")

    def gradual_unfreezing(self):

        for name, param in self.model.named_parameters():
            if 'embeddings' not in name and 'classification' not in name:
                param.detach_()
                param.requires_grad = False

            else:
                param.requires_grad = True

        full_parameters = sum(p.numel() for p in self.model.parameters())
        trained_parameters = sum(p.numel() for p in self.model.parameters() if p.requires_grad)

        logger.info(f"We will start by training {trained_parameters:3e} parameters out of {full_parameters:3e},"
                    f" i.e. {100 * trained_parameters / full_parameters:.2f}%")

        # We will unfreeze blocks regularly along the training: one block every `unfreezing_interval` step
        unfreezing_interval = int(len(self.dataset_splits.train_data_loader()) * self.num_epochs / (self.model.num_layers + 1))

        @self.trainer.on(Events.ITERATION_COMPLETED)
        def unfreeze_layer_if_needed(engine):
            if engine.state.iteration % unfreezing_interval == 0:
                # Which layer should we unfreeze now
                unfreezing_index = self.model.num_layers - (engine.state.iteration // unfreezing_interval)

                # Let's unfreeze it
                unfreezed = []
                for name, param in self.model.named_parameters():
                    if re.match(r"transformer\.[^\.]*\." + str(unfreezing_index) + r"\.", name):
                        unfreezed.append(name)
                        param.require_grad = True
                logger.info(f"Unfreezing block {unfreezing_index} with {unfreezed}")

    def discriminative_learning(self):

        logger.info("Using discriminative learning as adaptation strategy")
        # Build parameters groups by layer, numbered from the top ['1', '2', ..., '15']
        parameter_groups = []
        for i in range(self.model.num_layers):
            name_pattern = r"transformer\.[^\.]*\." + str(i) + r"\."
            group = {
                'name': str(self.model.num_layers - i),
                'params': [p for n, p in self.model.named_parameters() if re.match(name_pattern, n)]}
            parameter_groups.append(group)

        # Add the rest of the parameters (embeddings and classification layer) in a group labeled '0'
        name_pattern = r"transformer\.[^\.]*\.\d*\."
        group = {
            'name': '0',
            'params': [p for n, p in self.model.named_parameters() if not re.match(name_pattern, n)]}
        parameter_groups.append(group)

        # Sanity check that we still have the same number of parameters
        assert sum(p.numel() for g in parameter_groups for p in g['params']) \
               == sum(p.numel() for p in self.model.parameters())

        @self.trainer.on(Events.ITERATION_STARTED)
        def update_layer_learning_rates(engine):
            for param_group in self.optimizer.param_groups:
                layer_index = int(param_group["name"])
                param_group["lr"] = param_group["lr"] / (self.decreasing_factor ** layer_index)

        return parameter_groups

    def train(self):

        # if self.pretrained:
        #     self.load_pretrained_model()
        #
        # if self.adaptation == 'hard-freezing':
        #     self.freeze_params()
        # elif self.adaptation == 'gradual-unfreezing':
        #     self.gradual_unfreezing()
        # elif self.adaptation == 'discriminative-learning':
        #     parameter_groups = self.discriminative_learning()
        #     self.experiment_config.factories['optimizer'].kwargs['params'] = parameter_groups
        #     self.experiment_config.factories['optimizer'].param2config_key['params'] = parameter_groups
        # else:
        #     raise ValueError("Transfer NLP supports only hard freezing, gradual unfreezing and discriminative learning")
        #
        # self.optimizer = self.experiment_config.factories['optimizer'].create()
        # self.trainer.run(self.dataset_splits.train_data_loader(), max_epochs=self.num_epochs)
        raise NotImplementedError


@register_plugin
class MultiTaskTrainer(BaseIgniteTrainer):

    def __init__(self,
                 model: nn.Module,
                 dataset_splits: DatasetSplits,
                 loss: nn.Module,
                 optimizer: optim.Optimizer,
                 metrics: Dict[str, Metric],
                 device: str = None,
                 num_epochs: int = 1,
                 seed: int = None,
                 cuda: bool = None,
                 loss_accumulation_steps: int = 4,
                 scheduler: Any = None,
                 regularizer: RegularizerABC = None,
                 gradient_clipping: float = 1.0,
                 output_transform=None,
                 tensorboard_logs: str = None,
                 clf_loss_coef: float = 0.1,
                 lm_loss_coef: float = 0.9
                 ):

        super().__init__(
            model=model,
            dataset_splits=dataset_splits,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            device=device,
            num_epochs=num_epochs,
            seed=seed,
            cuda=cuda,
            loss_accumulation_steps=loss_accumulation_steps,
            scheduler=scheduler,
            regularizer=regularizer,
            gradient_clipping=gradient_clipping,
            output_transform=output_transform,
            tensorboard_logs=tensorboard_logs)
        self.clf_loss_coef = clf_loss_coef
        self.lm_loss_coef = lm_loss_coef
        RunningAverage(Accuracy(output_transform=lambda x: (x[0], x[1]))).attach(self.trainer, 'acc')

    def update_engine(self, engine, batch):
        self.model.train()
        batch = _prepare_batch(batch, device=self.device, non_blocking=False)
        lm_logits, clf_logits = self._forward(batch)
        loss_lm, loss_clf = self.loss(lm_logits=lm_logits, clf_logits=clf_logits, lm_labels=batch['x'], clf_labels=batch['y_target'])
        loss = (self.clf_loss_coef * loss_clf
                + self.lm_loss_coef * loss_lm) / self.loss_accumulation_steps

        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.gradient_clipping)

        if engine.state.iteration % self.loss_accumulation_steps == 0:
            self.optimizer.step()
            self.optimizer.zero_grad()

        return clf_logits, batch['y_target'], loss.item()

    def infer_engine(self, engine, batch):

        self.model.eval()
        with torch.no_grad():
            batch = _prepare_batch(batch, device=self.device, non_blocking=False)
            lm_logits, clf_logits = self._forward(batch)
            return clf_logits, batch['y_target']

    def create_supervised_evaluator(self):

        if self.device:
            self.model.to(self.device)

        engine = Engine(self.infer_engine)

        Accuracy().attach(engine, "accuracy")

        return engine

    def train(self):
        self.trainer.run(self.dataset_splits.train_data_loader(), max_epochs=self.num_epochs)



"""
This file contains all necessary plugins classes that the framework will use to let a user interact with custom models, data loaders, etc...
The Registry pattern used here is inspired from this post: https://realpython.com/primer-on-python-decorators/
"""
import logging
import os
import traceback
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Mapping, Type, Union

import toml
import yaml

logger = logging.getLogger(__name__)
REGISTRY = {}


def register_plugin(registrable: Any, alias: str = None):
    """
    Register a class, a function or a method to REGISTRY
    Args:
        registrable:
        alias:
    Returns:
    """
    alias = alias or registrable.__name__

    if alias in REGISTRY:
        raise ValueError(f"{alias} is already registered to registrable {REGISTRY[alias]}. Please select another name")

    REGISTRY[alias] = registrable
    return registrable


class InstantiationError(Exception):
    """
    An error happened while instantiating an experiment
    """

    def __init__(self, obj_name: str):
        """
        :param obj_name: The name of the object that couldn't be instantiated
        """
        self.obj_name: str = obj_name

    def __str__(self) -> str:
        return f"Failed to instantiate `{self.obj_name}`"


class CallableInstantiationError(InstantiationError):
    """
    An error happened while instantiating a callable
    """

    def __init__(self, obj_name: str, callable_name: str, arg_names: List[str]):
        """
        :param obj_name: The name of the object that couldn't be instantiated 
        :param callable_name: The callable's name that failed when called
        :param arg_names: The names of the arguments that were passed to the callable
        """
        super().__init__(obj_name)
        self.callable_name: str = callable_name
        self.arg_names: List[str] = arg_names

    def __str__(self) -> str:
        return f"{super().__str__()} calling `{self.callable_name}` using the arguments " + ', '.join(self.arg_names) + '; see exception raised above'


class LoopInConfigError(InstantiationError):
    """
    An object refers to itself
    """

    def __str__(self) -> str:
        return f"{super().__str__()} because it refers to itself"


class UnknownPluginException(InstantiationError):
    """
    A registrable has not been registred
    """
    def __init__(self, object_name: str, registrable: str):
        super().__init__(object_name)
        self.registrable: str = registrable

    def __str__(self) -> str:
        return f"{super().__str__()} because registrable `{self.registrable}` isn't registred"


class UnknownReferenceError(InstantiationError):
    """
    An object that is referenced does not exist
    """
    def __init__(self, object_name: str, reference_name: str):
        super().__init__(object_name)
        self.reference_name: str = reference_name

    def __str__(self) -> str:
        return f"{super().__str__()} because it reference to `{self.reference_name}` that doesn't exist"


class InstantiationImpossible(Exception):
    pass


class ObjectInstantiator(metaclass=ABCMeta):
    """
    An instantiator that knows how to instantiate 1 kind of object
    """

    def __init__(self):
        self.builder: ObjectBuilder = None

    def set_builder(self, builder: "ObjectBuilder") -> None:
        """
        Set the builder, for recursive puposes

        :param builder: The main builder to use when recursivity is needed
        :return: None
        """
        self.builder: ObjectBuilder = builder

    @abstractmethod
    def instantiate(self, config: Union[Dict, str, List], name: str) -> Any:
        """
        :param config: The config from which we want to instantiate the object
        :param name: The full name of the object to instantiate
        :return: The instantiated object
        """
        raise InstantiationImpossible


class ObjectBuilder:
    """
    The main builder class
    """
    def __init__(self, instantiators: List[ObjectInstantiator]):
        """
        :param instantiators: The ordered list of instantiators that can be used to instantiate the objects
        """
        self.instantiators: List[ObjectInstantiator] = instantiators
        for instantiator in instantiators:
            instantiator.set_builder(self)

    def instantiate(self, config: Union[Dict, str, List], name: str) -> Any:
        """
        build the object trying to use all the instantiators in a row, until one works

        :param config: The config of the object to instantiate
        :param name: The full name of this object
        :return: The instantiated object
        """

        for instantiator in self.instantiators:
            try:
                return instantiator.instantiate(config, name)
            except InstantiationImpossible:
                pass

        logger.info(f'instantiating "{name}" as a simple object, {config}')

        return config


class DictInstantiator(ObjectInstantiator):
    """
    Instantiate a dictionary
    """

    def instantiate(self, config: Union[Dict, str, List], name: str) -> Dict:
        """
        :param config: The config for the object to instantiate
        :param name: The full name of the object we try to instantiate
        :return: The resulting object
        :raise InstantiationImpossible: If the instantiator cannot handle this config
        """
        if not isinstance(config, dict):
            raise InstantiationImpossible()

        logger.info(f'instantiating "{name}" as a dictionary')

        return {
            key: self.builder.instantiate(value_config, f'{name}.{key}')
            for key, value_config in config.items()
        }


class ListInstantiator(ObjectInstantiator):
    """
    Instantiate a list
    """

    def instantiate(self, config: Union[Dict, str, List], name: str) -> List:
        """
        :param config: The config for the object to instantiate
        :param name: The full name of the object we try to instantiate
        :return: The resulting object
        :raise InstantiationImpossible: If the instantiator cannot handle this config
        """
        if not isinstance(config, list):
            raise InstantiationImpossible()

        logger.info(f'instantiating "{name}" as a list')

        return [
            self.builder.instantiate(value_config, f'{name}.{i}')
            for i, value_config in enumerate(config)
        ]


class FromMappingInstantiator(ObjectInstantiator):
    """
    Instantiate an object looking for its reference in a Mapping object
    """

    def __init__(self, mapping: Mapping[str, Any], mapping_name: str):
        """
        :param mapping: The mapping to use to look for references 
        :param mapping_name: The mapping name, for logs
        """
        self.env: Mapping[str, Any] = mapping
        self.mapping_name: str = mapping_name

        super().__init__()

    def instantiate(self, config: Union[Dict, str, List], name: str) -> Any:
        """
        :param config: The config for the object to instantiate
        :param name: The full name of the object we try to instantiate
        :return: The resulting object
        :raise InstantiationImpossible: If the instantiator cannot handle this config
        """

        if not isinstance(config, str) or not config.startswith('$'):
            raise InstantiationImpossible
        try:
            instance = self.env[config[1:]]
            logger.info(f'instantiating "{name}" using value {instance} from key {config} in {self.mapping_name}')
            return instance
        except KeyError:
            raise InstantiationImpossible


class FromEnvironmentVariableInstantiator(FromMappingInstantiator):
    """
    Instantiate an object from the environment variables
    """

    def __init__(self, env: Dict[str, Any]):
        """
        :param env: The dictionary of the environment variables
        """
        super().__init__(env, 'Environment')

        self.strings_to_replace: List[str, str] = [
            key
            for key, value in self.env.items()
            if isinstance(value, str) or isinstance(value, os.PathLike)
        ]
        self.strings_to_replace.sort(key=len, reverse=True)

    def instantiate(self, config: Union[Dict, str, List], name: str) -> Any:
        """
        :param config: The config for the object to instantiate
        :param name: The full name of the object we try to instantiate
        :return: The resulting object
        :raise InstantiationImpossible: If the instantiator cannot handle this config
        :raise UnknownReferenceError: If the config refers to an unknown object
        """
        try:
            return self.builder.instantiate(super().instantiate(config, name), f'{name}')
        except InstantiationImpossible:
            if not isinstance(config, str):
                raise InstantiationImpossible()

            v_upd: str = config
            for key in self.strings_to_replace:
                v_upd = v_upd.replace(f'${key}', str(self.env[key]))

            if v_upd.startswith('$'):
                raise UnknownReferenceError(name, config) from None

            logger.info(f'instantiating "{name}" using value {v_upd}')

            return v_upd


class CallableInstantiator(DictInstantiator):
    """
    Instantiate an object calling a registered callable
    """

    def instantiate(self, config: Union[Dict, str, List], name: str) -> Any:
        """
        :param config: The config for the object to instantiate
        :param name: The full name of the object we try to instantiate
        :return: The instantiated object
        :raise InstantiationImpossible: If the instantiator cannot handle this config
        :raise UnknownPluginException: If the callable is not registered
        :raise CallableInstantiationError: If an error occurred while calling the callable
        """
        if not isinstance(config, dict) or not '_name' in config:
            raise InstantiationImpossible()

        klass_name: str = config['_name']

        if klass_name not in REGISTRY:
            raise UnknownPluginException(object_name=name, registrable=klass_name)

        klass: Union[Type, Callable] = REGISTRY[klass_name]

        logger.info(f'instantiating "{name}" calling {klass_name}')

        param_instances: Dict[str, Any] = {
            key: self.builder.instantiate(value_config, f'{name}.{key}')
            for key, value_config in config.items()
            if key != '_name'
        }

        try:
            return klass(**param_instances)
        except Exception:
            raise CallableInstantiationError(obj_name=name, callable_name=klass_name, arg_names=list(param_instances.keys()))


class ExperimentConfig(Mapping[str, Any]):

    @staticmethod
    def load_experiment_config(experiment: Union[str, Path, Dict]) -> Dict:
        config = {}
        if isinstance(experiment, dict):
            config = dict(experiment)
        else:
            experiment_path = Path(str(experiment)).expanduser()
            with experiment_path.open() as f:
                if experiment_path.suffix in {'.json', '.yaml', '.yml'}:
                    config = yaml.safe_load(f)
                elif experiment_path.suffix in {'.toml'}:
                    config = toml.load(f)
                else:
                    raise ValueError("Only Dict, json, yaml and toml experiment files are supported")
        return config

    def __init__(self, experiment: Union[str, Path, Dict], **env):
        """
        :param experiment: the experiment config
        :param env: substitution variables, e.g. a HOME directory. generally use all caps.
        :return: the experiment
        """

        self.config: Dict[str, Any] = ExperimentConfig.load_experiment_config(experiment)

        self.builds_started: List[str] = []
        self.builders = [
            CallableInstantiator(),
            DictInstantiator(),
            ListInstantiator(),
            FromMappingInstantiator(REGISTRY, 'Registry'),
            FromMappingInstantiator(self, 'Experiment objects'),
            FromEnvironmentVariableInstantiator(env),
        ]

        self.builder: ObjectBuilder = ObjectBuilder(self.builders)

        self.experiment: Dict[str, Any] = {}

        for key, value_config in self.config.items():
            if key not in self.experiment:
                self.build(key)

    def _check_init(self):
        if self.experiment is None:
            raise ValueError('experiment config is not setup yet!')

    def build(self, key: str) -> Any:
        if key not in self.config:
            raise KeyError()
        if key in self.builds_started:
            raise LoopInConfigError(key)
        self.builds_started.append(key)
        self.experiment[key] = self.builder.instantiate(self.config[key], name=key)
        return self.experiment[key]

    # map-like methods
    def __getitem__(self, item):
        self._check_init()
        if item in self.experiment:
            return self.experiment[item]
        return self.build(item)

    def get(self, item, default=None):
        self._check_init()
        return self.experiment.get(item, default)

    def __iter__(self):
        self._check_init()
        return iter(self.experiment)

    def items(self):
        self._check_init()
        return self.experiment.items()

    def values(self):
        self._check_init()
        return self.experiment.values()

    def keys(self):
        self._check_init()
        return self.experiment.keys()

    def __setitem__(self, key, value):
        raise ValueError("cannot update experiment!")

    def __len__(self) -> int:
        return len(self.experiment)



from torch.utils.data import Dataset, DataLoader


class DatasetSplits:
    def __init__(self,
                 train_set: Dataset, train_batch_size: int,
                 val_set: Dataset, val_batch_size: int,
                 test_set: Dataset = None, test_batch_size: int = None):
        self.train_set: Dataset = train_set
        self.train_batch_size: int = train_batch_size

        self.val_set: Dataset = val_set
        self.val_batch_size: int = val_batch_size

        self.test_set: Dataset = test_set
        self.test_batch_size: int = test_batch_size

    def train_data_loader(self):
        return DataLoader(self.train_set, self.train_batch_size, shuffle=True)

    def val_data_loader(self):
        return DataLoader(self.val_set, self.val_batch_size, shuffle=False)

    def test_data_loader(self):
        return DataLoader(self.test_set, self.test_batch_size, shuffle=False)


# To use this class you will need to manually install pandas
class DataFrameDataset(Dataset):

    def __init__(self, df):
        self.df = df

    def __len__(self):
        return len(self.df)

    def __getitem__(self, item):
        row = self.df.iloc[item, :]
        return {col: row[col] for col in self.df.columns}


class DataProps:
    def __init__(self):
        self.input_dims: int = None
        self.output_dims: int = None



import logging

logger = logging.getLogger(__name__)

try:
    from transfer_nlp.plugins import metrics, regularizers, helpers, trainers
    logger.debug("Using trainers with Torch utilities")
except ImportError as e:
    logger.debug("Using trainers without Torch utilities")
